#######################################################################
# Tests for namedlist module.
#
# Copyright 2011-2014 True Blade Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Notes:
#  When moving to a newer version of unittest, check that the exceptions
# being caught have the expected text in them.
#
########################################################################

from namedlist import namedlist, namedtuple, FACTORY, NO_DEFAULT

import sys
import copy
import unittest
import collections
import unicodedata

_PY2 = sys.version_info[0] == 2
_PY3 = sys.version_info[0] == 3

# test both pickle and cPickle in 2.x, but just pickle in 3.x
import pickle
try:
    import cPickle
    pickle_modules = (pickle, cPickle)
except ImportError:
    pickle_modules = (pickle,)

# types used for pickle tests
TestNL0 = namedlist('TestNL0', '')
TestNL = namedlist('TestNL', 'x y z')

class TestNamedList(unittest.TestCase):
    def test_simple(self):
        Point = namedlist('Point', 'x y')
        p = Point(10, 20)
        self.assertEqual((p.x, p.y), (10, 20))
        self.assertEqual(p._asdict(), {'x':10, 'y':20})

        Point = namedlist('Point', 'x,y')
        p = Point(10, 20)
        self.assertEqual((p.x, p.y), (10, 20))
        self.assertEqual(p._asdict(), {'x':10, 'y':20})

        Point = namedlist('Point', 'x, y')
        p = Point(10, 20)
        self.assertEqual((p.x, p.y), (10, 20))
        self.assertEqual(p._asdict(), {'x':10, 'y':20})

        Point = namedlist('Point', ['x', 'y'])
        p = Point(10, 20)
        self.assertEqual((p.x, p.y), (10, 20))
        self.assertEqual(p._asdict(), {'x':10, 'y':20})

        self.assertEqual(Point(10, 11), Point(10, 11))
        self.assertNotEqual(Point(10, 11), Point(10, 12))

        self.assertEqual(vars(p), p._asdict())                              # verify that vars() works

    def test_unicode_identifiers(self):
        Point = namedlist(u'Point', u'x y')
        p = Point(10, 20)
        self.assertEqual((p.x, p.y), (10, 20))
        self.assertEqual(p._asdict(), {'x':10, 'y':20})

    def test_asdict_vars_ordered(self):
        Point = namedlist('Point', ['x', 'y'])
        p = Point(10, 20)

        # can't use unittest.skipIf in 2.6
        if sys.version_info[0] <= 2 and sys.version_info[1] <= 6:
            self.assertIsInstance(p.__dict__, dict)
        else:
            self.assertIsInstance(p.__dict__, collections.OrderedDict)

    def test_bad_name(self):
        self.assertRaises(ValueError, namedlist, 'Point*', 'x y')
        self.assertRaises(ValueError, namedlist, 'Point', '# y')
        self.assertRaises(ValueError, namedlist, 'Point', 'x 1y')
        self.assertRaises(ValueError, namedlist, 'Point', 'x y x')
        self.assertRaises(ValueError, namedlist, 'Point', 'x y for')
        self.assertRaises(ValueError, namedlist, 'Point', '_field')
        self.assertRaises(ValueError, namedlist, 'Point', [('', 0)])
        self.assertRaises(ValueError, namedlist, '', 'x y')
        self.assertRaises(ValueError, namedlist, 'Point', unicodedata.lookup('SUPERSCRIPT ONE'))

    def test_bad_defaults(self):
        # if specifying the defaults, must provide a 2-tuple
        self.assertRaises(ValueError, namedlist, 'Point', [('x', 3, 4)])
        self.assertRaises(ValueError, namedlist, 'Point', [('x',)])
        self.assertRaises(ValueError, namedlist, 'Point', [3])

    def test_empty(self):
        Point = namedlist('Point', '')
        self.assertEqual(len(Point()), 0)
        self.assertEqual(list(Point()), [])
        self.assertEqual(Point(), Point())
        self.assertEqual(Point()._asdict(), {})

        Point = namedlist('Point', '', 10)
        self.assertEqual(len(Point()), 0)
        self.assertEqual(Point(), Point())
        self.assertEqual(Point()._asdict(), {})

        Point = namedlist('Point', [])
        self.assertEqual(len(Point()), 0)
        self.assertEqual(Point(), Point())
        self.assertEqual(Point()._asdict(), {})

        Point = namedlist('Point', [], 10)
        self.assertEqual(len(Point()), 0)
        self.assertEqual(Point(), Point())
        self.assertEqual(Point()._asdict(), {})

    def test_list(self):
        Point = namedlist('Point', ['x', 'y'])
        p = Point(10, 20)
        self.assertEqual((p.x, p.y), (10, 20))
        self.assertEqual(p._asdict(), {'x':10, 'y':20})

        Point = namedlist('Point', ('x', 'y'))
        p = Point(10, 20)
        self.assertEqual((p.x, p.y), (10, 20))
        self.assertEqual(p._asdict(), {'x':10, 'y':20})

    def test_default(self):
        Point = namedlist('Point', 'x y z', 100)
        self.assertEqual(Point(), Point(100, 100, 100))
        self.assertEqual(Point(10), Point(10, 100, 100))
        self.assertEqual(Point(10, 20), Point(10, 20, 100))
        self.assertEqual(Point(10, 20, 30), Point(10, 20, 30))
        self.assertEqual(Point()._asdict(), {'x':100, 'y':100, 'z':100})

    def test_default_list(self):
        Point = namedlist('Point', 'x y z'.split(), 100)
        self.assertEqual(Point(), Point(100, 100, 100))
        self.assertEqual(Point(10), Point(10, 100, 100))
        self.assertEqual(Point(10, 20), Point(10, 20, 100))
        self.assertEqual(Point(10, 20, 30), Point(10, 20, 30))
        self.assertEqual(Point()._asdict(), {'x':100, 'y':100, 'z':100})

    def test_default_and_specified_default(self):
        Point = namedlist('Point', ['x', ('y', 10), ('z', 20)], 100)
        self.assertEqual(Point(), Point(100, 10, 20))
        self.assertEqual(Point(0), Point(0, 10, 20))
        self.assertEqual(Point(0, 1), Point(0, 1, 20))
        self.assertEqual(Point(0, 1, 2), Point(0, 1, 2))

        # default doesn't just have to apply to the last field
        Point = namedlist('Point', [('x', 0), 'y', ('z', 20)], 100)
        self.assertEqual(Point(), Point(0, 100, 20))

    def test_equality_inequality(self):
        Point = namedlist('Point', ['x', ('y', 10), ('z', 20)], 100)
        p0 = Point()
        p1 = Point(0)
        self.assertEqual(p0, Point())
        self.assertEqual(p0, Point(100, 10, 20))
        self.assertEqual(p1, Point(0, 10))
        self.assertEqual(Point(), p0)
        self.assertEqual(p0, p0)
        self.assertNotEqual(p0, p1)
        self.assertNotEqual(p0, 3)
        self.assertNotEqual(p0, None)
        self.assertNotEqual(p0, object())
        self.assertNotEqual(p0, Point('100'))
        self.assertNotEqual(p0, Point(100, 10, 21))

    def test_default_order(self):
        # with no default, can't have a field without a
        #  default follow fields with defaults
        self.assertRaises(ValueError, namedlist, 'Point',
                          ['x', ('y', 10), 'z'])

        # but with a default, you can
        Point = namedlist('Point', ['x', ('y', 10), 'z'], -1)
        self.assertEqual(Point(0), Point(0, 10, -1))
        self.assertEqual(Point(z=0), Point(-1, 10, 0))

    def test_repr(self):
        Point = namedlist('Point', 'x y z')
        p = Point(1, 2, 3)
        self.assertEqual(repr(p), 'Point(x=1, y=2, z=3)')
        self.assertEqual(str(p), 'Point(x=1, y=2, z=3)')

    def test_missing_argument(self):
        Point = namedlist('Point', ['x', 'y', ('z', 10)])
        self.assertEqual(Point(1, 2), Point(1, 2, 10))
        self.assertRaises(TypeError, Point, 1)

    def test_identity_of_defaults(self):
        default = object()
        Point = namedlist('Point', [('x', default)])
        # in 2.7 this should become assertIs
        self.assertTrue(Point().x is default)

        Point = namedlist('Point', 'x', default)
        # in 2.7 this should become assertIs
        self.assertTrue(Point().x is default)

    def test_writable(self):
        Point = namedlist('Point', ['x', ('y', 10), ('z', 20)], 100)
        p = Point(0)
        self.assertEqual((p.x, p.y, p.z), (0, 10, 20))
        p.x = -1
        self.assertEqual((p.x, p.y, p.z), (-1, 10, 20))
        p.y = -1
        self.assertEqual((p.x, p.y, p.z), (-1, -1, 20))
        p.z = None
        self.assertEqual((p.x, p.y, p.z), (-1, -1, None))

    def test_update_nothing(self):
        Point = namedlist('Point', ['x', ('y', 10), ('z', 20)], 100)

        a = Point(1)
        a._update()
        self.assertEqual((a.x, a.y, a.z), (1, 10, 20))

    def test_update_other_named_list(self):
        Point = namedlist('Point', ['x', ('y', 10), ('z', 20)], 100)

        a = Point(10)
        b = Point(100, 200, 300)
        a._update(b)

        self.assertEqual((a.x, a.y, a.z), (100, 200, 300))
        self.assertIsNot(a, b)

    def test_update_other_pairs(self):
        Point = namedlist('Point', ['x', ('y', 10), ('z', 20)], 100)

        a = Point(10)
        a._update([('y', 23), ('z', 42)])
        self.assertEqual((a.x, a.y, a.z), (10, 23, 42))

    def test_update_other_dict(self):
        Point = namedlist('Point', ['x', ('y', 10), ('z', 20)], 100)

        a = Point(10)
        a._update({'x': 32, 'y': 64})
        self.assertEqual((a.x, a.y, a.z), (32, 64, 20))

    def test_update_other_kwargs(self):
        Point = namedlist('Point', ['x', ('y', 10), ('z', 20)], 100)

        a = Point(10)
        a._update(x=60, z=90)
        self.assertEqual((a.x, a.y, a.z), (60, 10, 90))

    def test_update_other_pairs_kwargs(self):
        Point = namedlist('Point', ['x', ('y', 10), ('z', 20)], 100)

        a = Point(10)
        a._update([('y', 23)], z=42)
        self.assertEqual((a.x, a.y, a.z), (10, 23, 42))

    def test_update_other_dict_kwargs(self):
        Point = namedlist('Point', ['x', ('y', 10), ('z', 20)], 100)

        a = Point(10)
        a._update({'x': 32}, y=64)
        self.assertEqual((a.x, a.y, a.z), (32, 64, 20))

    def test_update_other_self_kwarg(self):
        # make sure 'self' and 'other' work as kwargs
        Point = namedlist('Point', 'self other')

        a = Point(1, 0)
        a._update(self=10, other=20)
        self.assertEqual((a.self, a.other), (10, 20))

    def test_replace_none(self):
        Point = namedlist('Point', 'x y z')
        a = Point(1, 2, 3)
        b = a._replace()
        self.assertIsNot(a, b)
        self.assertEqual((a.x, a.y, a.z), (1, 2, 3))
        self.assertEqual((b.x, b.y, b.z), (1, 2, 3))

    def test_replace_none(self):
        # make sure 'self' and 'other' work as kwargs
        Point = namedlist('Point', 'self other')
        a = Point(1, 2)
        b = a._replace(self=3, other=4)
        self.assertIsNot(a, b)
        self.assertEqual((a.self, a.other), (1, 2))
        self.assertEqual((b.self, b.other), (3, 4))

    def test_complex_defaults(self):
        Point = namedlist('Point', ['x', ('y', 10), ('z', 20)],
                          [1, 2, 3])
        p = Point()
        self.assertEqual((p.x, p.y, p.z), ([1, 2, 3], 10, 20))

        Point = namedlist('Point', [('x', [4, 5, 6]),
                                    ('y', 10),
                                    ('z', 20)])
        p = Point()
        self.assertEqual((p.x, p.y, p.z), ([4, 5, 6], 10, 20))

    def test_iteration(self):
        Point = namedlist('Point', ['x', ('y', 10), ('z', 20)],
                           [1, 2, 3])
        p = Point()
        self.assertEqual(len(p), 3)

        self.assertEqual(list(iter(p)), [[1, 2, 3], 10, 20])

        for expected, found in zip([[1, 2, 3], 10, 20], p):
            self.assertEqual(expected, found)

    def test_fields(self):
        Point = namedlist('Point', 'x y z')
        self.assertEqual(Point._fields, ('x', 'y', 'z'))
        self.assertIsInstance(Point._fields, tuple)

        Point = namedlist('Point', 'x y z', 100)
        self.assertEqual(Point._fields, ('x', 'y', 'z'))

        Point = namedlist('Point', [('x', 0), ('y', 0), ('z', 0)])
        self.assertEqual(Point._fields, ('x', 'y', 'z'))

        Point = namedlist('Point', '')
        self.assertEqual(Point._fields, ())
        self.assertIsInstance(Point._fields, tuple)

    def test_pickle(self):
        for p in (TestNL0(), TestNL(x=10, y=20, z=30)):
            for module in pickle_modules:
                for protocol in range(-1, module.HIGHEST_PROTOCOL + 1):
                    q = module.loads(module.dumps(p, protocol))
                    self.assertEqual(p, q)
                    self.assertEqual(p._fields, q._fields)
                    self.assertNotIn(b'OrderedDict', module.dumps(p, protocol))

    def test_type_has_same_name_as_field(self):
        Point = namedlist('Point',
                           ['Point', ('y', 10), ('z', 20)],
                           [1, 2, 3])
        p = Point()
        self.assertEqual(len(p), 3)
        self.assertEqual(p.Point, [1, 2, 3])

        Point = namedlist('Point', 'Point')
        p = Point(4)
        self.assertEqual(p.Point, 4)

        Point = namedlist('Point', 'x Point')
        p = Point(3, 4)
        self.assertEqual(p.Point, 4)

    def test_slots(self):
        Point = namedlist('Point', '')
        p = Point()
        # p.x = 3 raises AttributeError because of slots
        self.assertRaises(AttributeError, setattr, p, 'x', 3)

        Point = namedlist('Point', '', use_slots=True)
        p = Point()
        # p.x = 3 raises AttributeError because of slots
        self.assertRaises(AttributeError, setattr, p, 'x', 3)

    def test_no_slots(self):
        Point = namedlist('Point', '', use_slots=False)
        p = Point()
        # we should be able to create new attributes
        p.x = 3
        self.assertEqual(p.x, 3)

    def test_rename(self):
        Point = namedlist('Point', ('abc', 'def'), rename=True)
        self.assertEqual(Point._fields, ('abc', '_1'))

        Point = namedlist('Point', ('for', 'def'), rename=True)
        self.assertEqual(Point._fields, ('_0', '_1'))

        Point = namedlist('Point', 'a a b a b c', rename=True)
        self.assertEqual(Point._fields, ('a', '_1', 'b', '_3', '_4', 'c'))

        # nothing needs to be renamed, should still work with rename=True
        Point = namedlist('Point', 'x y z', rename=True)
        self.assertEqual(Point._fields, ('x', 'y', 'z'))

        Point = namedlist('Point', 'x y _z', rename=True)
        self.assertEqual(Point._fields, ('x', 'y', '_2'))

        # rename with defaults
        Point = namedlist('Point', [('', 1), ('', 2)], rename=True)
        p = Point()
        self.assertEqual(p._0, 1)
        self.assertEqual(p._1, 2)

    def test_type_begins_with_underscore(self):
        Point = namedlist('_Point', '')
        p = Point()

    def test_mapping(self):
        # use a regular dict so testing with 2.6 is still possible
        # do not make any assumptions about field order
        Point = namedlist('Point', {'x': 0, 'y': 100})
        p = Point()
        self.assertEqual(p.x, 0)
        self.assertEqual(p.y, 100)

        # in 2.7, test with an OrderedDict

    def test_NO_DEFAULT(self):
        # NO_DEFAULT is only really useful with we're using a mapping
        #  plus a default value. it's the only way to specify that
        #  some of the fields use the default.
        Point = namedlist('Point', {'x':0, 'y':NO_DEFAULT}, default=5)
        p = Point()
        self.assertEqual(p.x, 0)
        self.assertEqual(p.y, 5)

    def test_iterable(self):
        Point = namedlist('Point', iter(['x', 'y']))
        p = Point(1, 2)
        self.assertEqual(p.x, 1)
        self.assertEqual(p.y, 2)

    def test_single_field(self):
        X = namedlist('X', 'xyz')
        self.assertEqual(X._fields, ('xyz',))

    def test_repr_output(self):
        Point = namedlist('Point', 'a b')
        p = Point('0', 0)
        self.assertEqual(repr(p), "Point(a='0', b=0)")

    def test_mutable_defaults(self):
        # this behavior is unfortunate, but it should be tested anyway
        A = namedlist('A', [('x', [])])
        a = A()
        self.assertEqual(a.x, [])

        a.x.append(4)
        self.assertEqual(a.x, [4])

        b = A()
        self.assertEqual(b.x, [4])

    def test_factory_functions(self):
        A = namedlist('A', [('x', FACTORY(list))])
        a = A()
        self.assertEqual(a.x, [])

        a.x.append(4)
        self.assertEqual(a.x, [4])

        b = A()
        self.assertEqual(b.x, [])

    def test_factory_for_default(self):
        # make sure FACTORY works for the global default
        A = namedlist('A', 'x y', default=FACTORY(list))
        a = A()
        self.assertEqual(a.x, [])
        self.assertEqual(a.y, [])

        a.x.append(4)
        self.assertEqual(a.x, [4])
        a.y.append(4)
        self.assertEqual(a.y, [4])

        b = A()
        self.assertEqual(b.x, [])
        self.assertEqual(b.y, [])

        # mix and match FACTORY and a non-callable mutable default
        A = namedlist('A', [('x', []), 'y'], default=FACTORY(list))
        a = A()
        self.assertEqual(a.x, [])
        self.assertEqual(a.y, [])

        a.x.append(4)
        self.assertEqual(a.x, [4])
        a.y.append(4)
        self.assertEqual(a.y, [4])

        b = A()
        self.assertEqual(b.x, [4])
        self.assertEqual(b.y, [])

    def test_unhashable(self):
        Point = namedlist('Point', 'a b')
        p = Point(1, 2)
        self.assertRaises(TypeError, hash, p)

    def test_getitem(self):
        Point = namedlist('Point', 'a b')
        p = Point(1, 2)
        self.assertEqual((p[0], p[1]), (1, 2))
        self.assertEqual(list(p), [1, 2])
        self.assertRaises(IndexError, p.__getitem__, 2)

    def test_setitem(self):
        Point = namedlist('Point', 'a b')
        p = Point(1, 2)
        p[0] = 10
        self.assertEqual(list(p), [10, 2])
        p[1] = 20
        self.assertEqual(list(p), [10, 20])
        self.assertRaises(IndexError, p.__setitem__, 2, 3)

    def test_container(self):
        # I'm not sure there's much sense in this, but list is a container
        Point = namedlist('Point', 'a b')
        p = Point(1, 2)
        self.assertIn(2, p)

    def test_ABC(self):
        Point = namedlist('Point', 'a b c')
        p = Point(1, 2, 2)
        self.assertIsInstance(p, collections.Container)
        self.assertIsInstance(p, collections.Iterable)
        self.assertIsInstance(p, collections.Sized)
        self.assertIsInstance(p, collections.Sequence)

        self.assertEqual(list(reversed(p)), [2, 2, 1])
        self.assertEqual(p.count(0), 0)
        self.assertEqual(p.count(2), 2)
        self.assertRaises(ValueError, p.index, 0)
        self.assertEqual(p.index(2), 1)
        self.assertEqual(p.index(1), 0)

        A = namedlist('A', 'a b c d e f g h i j')
        a = A(0, 0, 0, 0, 0, 0, 0, 0, 0, 0)
        self.assertEqual(a.index(0, 1), 1)
        self.assertEqual(a.index(0, 5), 5)
        self.assertEqual(a.index(0, 1, 3), 1)
        self.assertEqual(a.index(0, 5, 12), 5)
        self.assertRaises(ValueError, a.index, 0, 12)

    def test_docstring(self):
        Point = namedlist('Point', '')
        self.assertEqual(Point.__doc__, 'Point()')

        Point = namedlist('Point', 'dx')
        self.assertEqual(Point.__doc__, 'Point(dx)')

        Point = namedlist('Point', 'x')
        self.assertEqual(Point.__doc__, 'Point(x)')

        Point = namedlist('Point', 'dx dy, dz')
        self.assertEqual(Point.__doc__, 'Point(dx, dy, dz)')

        Point = namedlist('Point', 'dx dy dz', default=10)
        self.assertEqual(Point.__doc__, 'Point(dx=10, dy=10, dz=10)')

        Point = namedlist('Point', 'dx, dy, dz', default=FACTORY(10))
        self.assertEqual(Point.__doc__, 'Point(dx=FACTORY(10), dy=FACTORY(10), dz=FACTORY(10))')

        Point = namedlist('Point', ['dx', 'dy', ('dz', 11.0)], default=10)
        self.assertEqual(Point.__doc__, 'Point(dx=10, dy=10, dz=11.0)')

        Point = namedlist('Point', ['dx', 'dy', ('dz', 11.0)], default=FACTORY(list))
        if _PY2:
            list_repr = "<type 'list'>"
        else:
            list_repr = "<class 'list'>"
        self.assertEqual(Point.__doc__, "Point(dx=FACTORY({0}), dy=FACTORY({0}), dz=11.0)".format(list_repr))

        Point = namedlist('Point', ['dx', 'dy', ('dz', FACTORY(11.0))], default=[])
        self.assertEqual(Point.__doc__, 'Point(dx=[], dy=[], dz=FACTORY(11.0))')

    def test_slice(self):
        Point = namedlist('Point', 'x y z color')
        values = [3, 5, -12, 'red']
        p = Point(*values)
        self.assertEqual(values[0:-1], p[0:-1])
        self.assertEqual(values[:3], p[:3])
        self.assertEqual(values[4:1:-1], p[4:1:-1])



TestNT = namedtuple('TestNT', 'x y z')    # type used for pickle tests

class TestNamedTuple(unittest.TestCase):

    def test_unicode_identifiers(self):
        Point = namedtuple(u'Point', u'x y')
        p = Point(10, 20)
        self.assertEqual((p.x, p.y), (10, 20))
        self.assertEqual(p._asdict(), {'x':10, 'y':20})

    def test_factory(self):
        Point = namedtuple('Point', 'x y')
        self.assertEqual(Point.__name__, 'Point')
        self.assertEqual(Point.__slots__, ())
        self.assertEqual(Point.__module__, __name__)
        self.assertEqual(Point.__getitem__, tuple.__getitem__)
        self.assertEqual(Point._fields, ('x', 'y'))

        self.assertRaises(ValueError, namedtuple, 'abc%', 'efg ghi')       # type has non-alpha char
        self.assertRaises(ValueError, namedtuple, 'class', 'efg ghi')      # type has keyword
        self.assertRaises(ValueError, namedtuple, '9abc', 'efg ghi')       # type starts with digit

        self.assertRaises(ValueError, namedtuple, 'abc', 'efg g%hi')       # field with non-alpha char
        self.assertRaises(ValueError, namedtuple, 'abc', 'abc class')      # field has keyword
        self.assertRaises(ValueError, namedtuple, 'abc', '8efg 9ghi')      # field starts with digit
        self.assertRaises(ValueError, namedtuple, 'abc', '_efg ghi')       # field with leading underscore
        self.assertRaises(ValueError, namedtuple, 'abc', 'efg efg ghi')    # duplicate field

        namedtuple('Point0', 'x1 y2')   # Verify that numbers are allowed in names
        namedtuple('_', 'a b c')        # Test leading underscores in a typename

        nt = namedtuple('nt', 'the quick brown fox')                       # check unicode input
        self.assertNotIn("u'", repr(nt._fields))
        nt = namedtuple('nt', ('the', 'quick'))                           # check unicode input
        self.assertNotIn("u'", repr(nt._fields))

        self.assertRaises(TypeError, Point._make, [11])                     # catch too few args
        self.assertRaises(TypeError, Point._make, [11, 22, 33])             # catch too many args

    @unittest.skipIf(sys.flags.optimize >= 2,
                     "Docstrings are omitted with -O2 and above")
    def test_factory_doc_attr(self):
        Point = namedtuple('Point', 'x y')
        self.assertEqual(Point.__doc__, 'Point(x, y)')

    def test_name_fixer(self):
        for spec, renamed in [
            [('efg', 'g%hi'),  ('efg', '_1')],                              # field with non-alpha char
            [('abc', 'class'), ('abc', '_1')],                              # field has keyword
            [('8efg', '9ghi'), ('_0', '_1')],                               # field starts with digit
            [('abc', '_efg'), ('abc', '_1')],                               # field with leading underscore
            [('abc', 'efg', 'efg', 'ghi'), ('abc', 'efg', '_2', 'ghi')],    # duplicate field
            [('abc', '', 'x'), ('abc', '_1', 'x')],                         # fieldname is a space
        ]:
            self.assertEqual(namedtuple('NT', spec, rename=True)._fields, renamed)

    def test_instance(self):
        Point = namedtuple('Point', 'x y')
        p = Point(11, 22)
        self.assertEqual(p, Point(x=11, y=22))
        self.assertEqual(p, Point(11, y=22))
        self.assertEqual(p, Point(y=22, x=11))
        self.assertEqual(p, Point(*(11, 22)))
        self.assertEqual(p, Point(**dict(x=11, y=22)))
        self.assertRaises(TypeError, Point, 1)                              # too few args
        self.assertRaises(TypeError, Point, 1, 2, 3)                        # too many args
        self.assertRaises(TypeError, eval, 'Point(XXX=1, y=2)', locals())   # wrong keyword argument
        self.assertRaises(TypeError, eval, 'Point(x=1)', locals())          # missing keyword argument
        self.assertEqual(repr(p), 'Point(x=11, y=22)')
        self.assertNotIn('__weakref__', dir(p))
        self.assertEqual(p, Point._make([11, 22]))                          # test _make classmethod
        self.assertEqual(p._fields, ('x', 'y'))                             # test _fields attribute
        self.assertEqual(p._replace(x=1), (1, 22))                          # test _replace method
        self.assertEqual(p._asdict(), dict(x=11, y=22))                     # test _asdict method
        self.assertEqual(vars(p), p._asdict())                              # verify that vars() works

        try:
            p._replace(x=1, error=2)
        except ValueError:
            pass
        else:
            self._fail('Did not detect an incorrect fieldname')

        # verify that field string can have commas
        Point = namedtuple('Point', 'x, y')
        p = Point(x=11, y=22)
        self.assertEqual(repr(p), 'Point(x=11, y=22)')

        # verify that fieldspec can be a non-string sequence
        Point = namedtuple('Point', ('x', 'y'))
        p = Point(x=11, y=22)
        self.assertEqual(repr(p), 'Point(x=11, y=22)')

    def test_tupleness(self):
        Point = namedtuple('Point', 'x y')
        p = Point(11, 22)

        self.assertIsInstance(p, tuple)
        self.assertEqual(p, (11, 22))                                       # matches a real tuple
        self.assertEqual(tuple(p), (11, 22))                                # coercable to a real tuple
        self.assertEqual(list(p), [11, 22])                                 # coercable to a list
        self.assertEqual(max(p), 22)                                        # iterable
        self.assertEqual(max(*p), 22)                                       # star-able
        x, y = p
        self.assertEqual(p, (x, y))                                         # unpacks like a tuple
        self.assertEqual((p[0], p[1]), (11, 22))                            # indexable like a tuple
        self.assertRaises(IndexError, p.__getitem__, 3)

        self.assertEqual(p.x, x)
        self.assertEqual(p.y, y)
        self.assertRaises(AttributeError, eval, 'p.z', locals())

    def test_odd_sizes(self):
        Zero = namedtuple('Zero', '')
        self.assertEqual(Zero(), ())
        self.assertEqual(Zero._make([]), ())
        self.assertEqual(repr(Zero()), 'Zero()')
        self.assertEqual(Zero()._asdict(), {})
        self.assertEqual(Zero()._fields, ())

        Dot = namedtuple('Dot', 'd')
        self.assertEqual(Dot(1), (1,))
        self.assertEqual(Dot._make([1]), (1,))
        self.assertEqual(Dot(1).d, 1)
        self.assertEqual(repr(Dot(1)), 'Dot(d=1)')
        self.assertEqual(Dot(1)._asdict(), {'d':1})
        self.assertEqual(Dot(1)._replace(d=999), (999,))
        self.assertEqual(Dot(1)._fields, ('d',))

        # n = 5000
        n = 254 # SyntaxError: more than 255 arguments:
        import string, random
        names = list(set(''.join([random.choice(string.ascii_letters)
                                  for j in range(10)]) for i in range(n)))
        n = len(names)
        Big = namedtuple('Big', names)
        b = Big(*range(n))
        self.assertEqual(b, tuple(range(n)))
        self.assertEqual(Big._make(range(n)), tuple(range(n)))
        for pos, name in enumerate(names):
            self.assertEqual(getattr(b, name), pos)
        repr(b)                                 # make sure repr() doesn't blow-up
        d = b._asdict()
        d_expected = dict(zip(names, range(n)))
        self.assertEqual(d, d_expected)
        b2 = b._replace(**dict([(names[1], 999),(names[-5], 42)]))
        b2_expected = list(range(n))
        b2_expected[1] = 999
        b2_expected[-5] = 42
        self.assertEqual(b2, tuple(b2_expected))
        self.assertEqual(b._fields, tuple(names))

    def test_pickle(self):
        p = TestNT(x=10, y=20, z=30)
        for module in (pickle,):
            loads = getattr(module, 'loads')
            dumps = getattr(module, 'dumps')
            for protocol in -1, 0, 1, 2:
                q = loads(dumps(p, protocol))
                self.assertEqual(p, q)
                self.assertEqual(p._fields, q._fields)
                self.assertNotIn(b'OrderedDict', dumps(p, protocol))

    def test_copy(self):
        p = TestNT(x=10, y=20, z=30)
        for copier in copy.copy, copy.deepcopy:
            q = copier(p)
            self.assertEqual(p, q)
            self.assertEqual(p._fields, q._fields)

    def test_name_conflicts(self):
        # Some names like "self", "cls", "tuple", "itemgetter", and "property"
        # failed when used as field names.  Test to make sure these now work.
        T = namedtuple('T', 'itemgetter property self cls tuple')
        t = T(1, 2, 3, 4, 5)
        self.assertEqual(t, (1,2,3,4,5))
        newt = t._replace(itemgetter=10, property=20, self=30, cls=40, tuple=50)
        self.assertEqual(newt, (10,20,30,40,50))


    def test_repr(self):
        A = namedtuple('A', 'x')
        self.assertEqual(repr(A(1)), 'A(x=1)')
        # repr should show the name of the subclass
        class B(A):
            pass
        self.assertEqual(repr(B(1)), 'B(x=1)')


class TestAll(unittest.TestCase):
    def test_all(self):
        import namedlist

        # check that __all__ in the module contains everything that should be
        #  public, and only those symbols
        all = set(namedlist.__all__)

        # check that things in __all__ only appear once
        self.assertEqual(len(all), len(namedlist.__all__),
                         'some symbols appear more than once in __all__')

        # get the list of public symbols
        found = set(name for name in dir(namedlist) if not name.startswith('_'))

        # make sure it matches __all__
        self.assertEqual(all, found)



# 2.6 is missing some unittest.TestCase members. Add
#  trivial implementations for them.
def _assertIsInstance(self, obj, cls):
    self.assertTrue(isinstance(obj, cls))

def _assertIn(self, obj, iterable):
    self.assertTrue(obj in iterable)

def _assertNotIn(self, obj, iterable):
    self.assertTrue(not obj in iterable)

def _add_unittest_methods(cls):
    for name, fn in [('assertIsInstance', _assertIsInstance),
                     ('assertIn', _assertIn),
                     ('assertNotIn', _assertNotIn)]:
        if not hasattr(cls, name):
            setattr(cls, name, fn)

_add_unittest_methods(TestNamedList)

unittest.main()



#######################################################################
# namedlist is similar to collections.namedtuple, but supports default
#  values and is writable.  Also contains an implementation of
#  namedtuple, which is the same as collections.namedtuple supporting
#  defaults.
#
# Copyright 2011-2014 True Blade Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Notes:
#  See http://code.activestate.com/recipes/576555/ for a similar
# concept.
#
########################################################################

__all__ = ['namedlist', 'namedtuple', 'NO_DEFAULT', 'FACTORY']

# All of this hassle with ast is solely to provide a decent __init__
#  function, that takes all of the right arguments and defaults. But
#  it's worth it to get all of the normal python error messages.
# For other functions, like __repr__, we don't bother. __init__ is
#  the only function where we really need the argument processing,
#  because __init__ is the only function whose signature will vary
#  per class.

import ast as _ast
import sys as _sys
import copy as _copy
import operator as _operator
import itertools as _itertools
from keyword import iskeyword as _iskeyword
import collections as _collections
import abc as _abc

_PY2 = _sys.version_info[0] == 2
_PY3 = _sys.version_info[0] == 3

try:
    _OrderedDict = _collections.OrderedDict
except AttributeError:
    _OrderedDict = None


if _PY2:
    _basestring = basestring
    _iteritems = lambda d, **kw: iter(d.iteritems(**kw))
else:
    _basestring = str
    _iteritems = lambda d, **kw: iter(d.items(**kw))


NO_DEFAULT = object()

# Wrapper around a callable. Used to specify a factory function instead
#  of a plain default value.
class FACTORY(object):
    def __init__(self, callable):
        self._callable = callable

    def __repr__(self):
        return 'FACTORY({0!r})'.format(self._callable)


########################################################################
# Keep track of fields, both with and without defaults.
class _Fields(object):
    default_not_specified = object()

    def __init__(self, default):
        self.default = default
        self.with_defaults = []        # List of (field_name, default).
        self.without_defaults = []     # List of field_name.

    def add(self, field_name, default):
        if default is self.default_not_specified:
            if self.default is NO_DEFAULT:
                # No default. There can't be any defaults already specified.
                if len(self.with_defaults) != 0:
                    raise ValueError('field {0} without a default follows fields '
                                     'with defaults'.format(field_name))
                self.without_defaults.append(field_name)
            else:
                self.add(field_name, self.default)
        else:
            if default is NO_DEFAULT:
                self.add(field_name, self.default_not_specified)
            else:
                self.with_defaults.append((field_name, default))


########################################################################
# Validate and possibly sanitize the field and type names.
class _NameChecker(object):
    def __init__(self, typename):
        self.seen_fields = set()
        self._check_common(typename, 'Type')

    def check_field_name(self, fieldname, rename, idx):
        try:
            self._check_common(fieldname, 'Field')
            self._check_specific_to_fields(fieldname)
        except ValueError as ex:
            if rename:
                return '_' + str(idx)
            else:
                raise

        self.seen_fields.add(fieldname)
        return fieldname

    def _check_common(self, name, type_of_name):
        # tests that are common to both field names and the type name
        if len(name) == 0:
            raise ValueError('{0} names cannot be zero '
                             'length: {1!r}'.format(type_of_name, name))
        if _PY2:
            if not all(c.isalnum() or c=='_' for c in name):
                raise ValueError('{0} names can only contain '
                                 'alphanumeric characters and underscores: '
                                 '{1!r}'.format(type_of_name, name))
            if name[0].isdigit():
                raise ValueError('{0} names cannot start with a '
                                 'number: {1!r}'.format(type_of_name, name))
        else:
            if not name.isidentifier():
                raise ValueError('{0} names names must be valid '
                                 'identifiers: {1!r}'.format(type_of_name, name))
        if _iskeyword(name):
            raise ValueError('{0} names cannot be a keyword: '
                             '{1!r}'.format(type_of_name, name))

    def _check_specific_to_fields(self, name):
        # these tests don't apply for the typename, just the fieldnames
        if name in self.seen_fields:
            raise ValueError('Encountered duplicate field name: '
                             '{0!r}'.format(name))

        if name.startswith('_'):
            raise ValueError('Field names cannot start with an underscore: '
                             '{0!r}'.format(name))


########################################################################
# Returns a function with name 'name', that calls another function 'chain_fn'
# This is used to create the __init__ function with the right argument names and defaults, that
#  calls into _init to do the real work.
# The new function takes args as arguments, with defaults as given.
def _make_fn(name, chain_fn, args, defaults):
    args_with_self = ['_self'] + list(args)
    arguments = [_ast.Name(id=arg, ctx=_ast.Load()) for arg in args_with_self]
    defs = [_ast.Name(id='_def{0}'.format(idx), ctx=_ast.Load()) for idx, _ in enumerate(defaults)]
    if _PY2:
        parameters = _ast.arguments(args=[_ast.Name(id=arg, ctx=_ast.Param()) for arg in args_with_self],
                                    defaults=defs)
    else:
        parameters = _ast.arguments(args=[_ast.arg(arg=arg) for arg in args_with_self],
                                    kwonlyargs=[],
                                    defaults=defs,
                                    kw_defaults=[])
    module_node = _ast.Module(body=[_ast.FunctionDef(name=name,
                                                     args=parameters,
                                                     body=[_ast.Return(value=_ast.Call(func=_ast.Name(id='_chain', ctx=_ast.Load()),
                                                                                       args=arguments,
                                                                                       keywords=[]))],
                                                     decorator_list=[])])
    module_node = _ast.fix_missing_locations(module_node)

    # compile the ast
    code = compile(module_node, '<string>', 'exec')

    # and eval it in the right context
    globals_ = {'_chain': chain_fn}
    locals_ = dict(('_def{0}'.format(idx), value) for idx, value in enumerate(defaults))
    eval(code, globals_, locals_)

    # extract our function from the newly created module
    return locals_[name]


########################################################################
# Produce a docstring for the class.

def _field_name_with_default(name, default):
    if default is NO_DEFAULT:
        return name
    return '{0}={1!r}'.format(name, default)

def _build_docstring(typename, fields, defaults):
    # We can use NO_DEFAULT as a sentinel here, becuase it will never be
    #  present in defaults. By this point, it has been removed and replaced
    #  with actual default values.

    # The defaults make this a little tricky. Append a sentinel in
    #  front of defaults until it's the same length as fields. The
    #  sentinel value is used in _name_with_default
    defaults = [NO_DEFAULT] * (len(fields) - len(defaults)) + defaults
    return '{0}({1})'.format(typename, ', '.join(_field_name_with_default(name, default)
                                                 for name, default in zip(fields, defaults)))


########################################################################
# Given the typename, fields_names, default, and the rename flag,
#  return a tuple of fields and a list of defaults.
def _fields_and_defaults(typename, field_names, default, rename):
    # field_names must be a string or an iterable, consisting of fieldname
    #  strings or 2-tuples. Each 2-tuple is of the form (fieldname,
    #  default).

    # Keeps track of the fields we're adding, with their defaults.
    fields = _Fields(default)

    # Validates field and type names.
    name_checker = _NameChecker(typename)

    if isinstance(field_names, _basestring):
        # No per-field defaults. So it's like a collections.namedtuple,
        #  but with a possible default value.
        field_names = field_names.replace(',', ' ').split()

    # If field_names is a Mapping, change it to return the
    #  (field_name, default) pairs, as if it were a list.
    if isinstance(field_names, _collections.Mapping):
        field_names = field_names.items()

    # Parse and validate the field names.

    # field_names is now an iterable. Walk through it,
    # sanitizing as needed, and add to fields.

    for idx, field_name in enumerate(field_names):
        if isinstance(field_name, _basestring):
            default = fields.default_not_specified
        else:
            try:
                if len(field_name) != 2:
                    raise ValueError('field_name must be a 2-tuple: '
                                     '{0!r}'.format(field_name))
            except TypeError:
                # field_name doesn't have a __len__.
                raise ValueError('field_name must be a 2-tuple: '
                                 '{0!r}'.format(field_name))
            default = field_name[1]
            field_name = field_name[0]

        # Okay: now we have the field_name and the default value (if any).
        # Validate the name, and add the field.
        # Convert field_name to str for python 2.x
        fields.add(name_checker.check_field_name(str(field_name), rename, idx), default)

    return (tuple(fields.without_defaults + [name for name, default in
                                             fields.with_defaults]),
            [default for _, default in fields.with_defaults])

########################################################################
# Common member functions for the generated classes.

def _repr(self):
    return '{0}({1})'.format(self.__class__.__name__, ', '.join('{0}={1!r}'.format(name, getattr(self, name)) for name in self._fields))

def _asdict(self):
    # In 2.6, return a dict.
    # Otherwise, return an OrderedDict
    t = _OrderedDict if _OrderedDict is not None else dict
    return t(zip(self._fields, self))

# Set up methods and fields shared by namedlist and namedtuple
def _common_fields(fields, docstr):
    type_dict = {'__repr__': _repr,
                 '__dict__': property(_asdict),
                 '__doc__': docstr,
                 '_asdict': _asdict,
                 '_fields': fields}

    # See collections.namedtuple for a description of
    #  what's happening here.
    # _getframe(2) instead of 1, because we're now inside
    #  another function.
    try:
        type_dict['__module__'] = _sys._getframe(2).f_globals.get('__name__', '__main__')
    except (AttributeError, ValueError):
        pass
    return type_dict


########################################################################
# namedlist methods

# The function that __init__ calls to do the actual work.
def _nl_init(self, *args):
    # sets all of the fields to their passed in values
    for fieldname, value in _get_values(self._fields, args):
        setattr(self, fieldname, value)

def _nl_eq(self, other):
    return isinstance(other, self.__class__) and all(getattr(self, name) == getattr(other, name) for name in self._fields)

def _nl_ne(self, other):
    return not _nl_eq(self, other)

def _nl_len(self):
    return len(self._fields)

def _nl_getstate(self):
    return tuple(getattr(self, fieldname) for fieldname in self._fields)

def _nl_setstate(self, state):
    for fieldname, value in zip(self._fields, state):
        setattr(self, fieldname, value)

def _nl_getitem(self, idx):
    if isinstance(idx, slice):
        # this isn't super-efficient, but works
        return [getattr(self, self._fields[i]) for i in range(*idx.indices(len(self._fields)))]

    return getattr(self, self._fields[idx])

def _nl_setitem(self, idx, value):
    return setattr(self, self._fields[idx], value)

def _nl_iter(self):
    return (getattr(self, fieldname) for fieldname in self._fields)

def _nl_count(self, value):
    return sum(1 for v in iter(self) if v == value)

def _nl_index(self, value, start=NO_DEFAULT, stop=NO_DEFAULT):
    # not the most efficient way to implement this, but it will work
    l = list(self)
    if start is NO_DEFAULT and stop is NO_DEFAULT:
        return l.index(value)
    if stop is NO_DEFAULT:
        return l.index(value, start)
    return l.index(value, start, stop)

def _nl_update(_self, _other=None, **kwds):
    if isinstance(_other, type(_self)):
        _other = zip(_self._fields, _other)
    elif isinstance(_other, _collections.Mapping):
        tmp = []
        for field_name in _self._fields:
            try:
                other_value = _other[field_name]
            except KeyError:
                pass
            else:
                tmp.append((field_name, other_value))
        _other = tmp
    elif _other is None:
        _other = []

    chained = _itertools.chain(_other, (x for x in _iteritems(kwds)
                                        if x[0] in _self._fields))
    for key, value in chained:
        setattr(_self, key, value)

def _nl_replace(_self, **kwds):
    # make a shallow copy, then mutate it
    other = _copy.copy(_self)
    for key, value in _iteritems(kwds):
        setattr(other, key, value)
    return other

########################################################################
# The actual namedlist factory function.
def namedlist(typename, field_names, default=NO_DEFAULT, rename=False,
              use_slots=True):
    typename = str(typename) # for python 2.x
    fields, defaults = _fields_and_defaults(typename, field_names, default, rename)

    type_dict = {'__init__': _make_fn('__init__', _nl_init, fields, defaults),
                 '__eq__': _nl_eq,
                 '__ne__': _nl_ne,
                 '__len__': _nl_len,
                 '__getstate__': _nl_getstate,
                 '__setstate__': _nl_setstate,
                 '__getitem__': _nl_getitem,
                 '__setitem__': _nl_setitem,
                 '__iter__': _nl_iter,
                 '__hash__': None,
                 'count': _nl_count,
                 'index': _nl_index,
                 '_update': _nl_update,
                 '_replace': _nl_replace}
    type_dict.update(_common_fields(fields, _build_docstring(typename, fields, defaults)))

    if use_slots:
        type_dict['__slots__'] = fields

    # Create the new type object.
    t = type(typename, (object,), type_dict)

    # Register its ABC's
    _collections.Sequence.register(t)

    # And return it.
    return t


########################################################################
# namedtuple methods
def _nt_new(cls, *args):
    # sets all of the fields to their passed in values
    assert len(args) == len(cls._fields)
    values = [value for _, value in _get_values(cls._fields, args)]
    return tuple.__new__(cls, values)

def _nt_replace(_self, **kwds):
    result = _self._make(map(kwds.pop, _self._fields, _self))
    if kwds:
        raise ValueError('Got unexpected field names: %r' % list(kwds))
    return result

def _nt_make(cls, iterable, new=tuple.__new__):
    result = new(cls, iterable)
    if len(result) != len(cls._fields):
        raise TypeError('Expected {0} arguments, got {1}'.format(len(cls._fields), len(result)))
    return result

def _nt_getnewargs(self):
    'Return self as a plain tuple.  Used by copy and pickle.'
    return tuple(self)

def _nt_getstate(self):
    'Exclude the OrderedDict from pickling'
    return None

def _get_values(fields, args):
    # Returns [(fieldname, value)]. If the value is a FACTORY, call it.
    assert len(fields) == len(args)
    return [(fieldname, (value._callable() if isinstance(value, FACTORY) else value))
            for fieldname, value in zip(fields, args)]

########################################################################
# The actual namedtuple factory function.
def namedtuple(typename, field_names, default=NO_DEFAULT, rename=False):
    typename = str(typename) # for python 2.x
    fields, defaults = _fields_and_defaults(typename, field_names, default, rename)

    type_dict = {'__new__': _make_fn('__new__', _nt_new, fields, defaults),
                 '__getnewargs__': _nt_getnewargs,
                 '__getstate__': _nt_getstate,
                 '_replace': _nt_replace,
                 '_make': classmethod(_nt_make),
                 '__slots__': ()}
    type_dict.update(_common_fields(fields, _build_docstring(typename, fields, defaults)))

    # Create each field property.
    for idx, field in enumerate(fields):
        type_dict[field] = property(_operator.itemgetter(idx),
                                    doc='Alias for field number {0}'.format(idx))

    # Create the new type object.
    return type(typename, (tuple,), type_dict)
